from dataclasses import dataclass, field
from typing import Callable
import heapq

__all__ = ['subscribe', 'unsubscribe', 'publish', 'run_until', 'empty']

_events = []
_topics = {}


def subscribe(topic: str, handler: Callable):
    """
    Subscribe a handler function to a topic.

    Parameters
    ----------
    topic : str
        The topic to subscribe to.
    handler : Callable
        The handler function to be called when events are published to the topic.
    """
    if topic not in _topics:
        _topics[topic] = []
    _topics[topic].append(handler)


def unsubscribe(topic: str, handler: Callable):
    """
    Unsubscribe a handler function from a topic.

    Parameters
    ----------
    topic : str
        The topic to unsubscribe from.
    handler : Callable
        The handler function to be removed from the topic.
    """
    if topic in _topics:
        _topics[topic].remove(handler)


def publish(topic: str, delivery_time: float, *args, **kwargs):
    """
    Publish an event to a topic with a specific delivery time.

    Parameters
    ----------
    topic : str
        The topic to publish the event to.
    delivery_time : float
        The time at which the event should be delivered.
    *args : list
        Variable-length argument list to be passed to the event handlers.
    **kwargs : dict
        Arbitrary keyword arguments to be passed to the event handlers.

    Raises
    ------
    KeyError
        If the specified topic does not exist.
    """
    if topic not in _topics:
        raise KeyError(f"Topic '{topic}' does not exist.")

    event = (delivery_time, topic, args, kwargs)
    heapq.heappush(_events, event)


def run_until(target_time: float):
    """
    Process events in the queue until the target time is reached.

    Parameters
    ----------
    target_time : float
        The time until which events should be processed.
    """
    while _events and _events[0][0] <= target_time:
        event_time, topic, args, kwargs = heapq.heappop(_events)

        if topic in _topics:
            for handler in _topics[topic]:
                handler(*args, **kwargs)


def empty() -> bool:
    """
    Check if there are any undelivered events in the queue.

    Returns
    -------
    bool
        True if there are undelivered events in the queue, False otherwise.
    """
    return not bool(_events)
